@@ -155,20 +155,23 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 		if ctx.Err() != nil {
 			return nil, ctx.Err()
 		}
-		pkg, err := snapshot.PackageForFile(ctx, fh.URI(), source.TypecheckFull, source.WidestPackage)
-		if err != nil {
-			return nil, err
-		}
 
-		pkgDiagnostics, err := snapshot.DiagnosePackage(ctx, pkg)
+		// Type-check the package and also run analysis,
+		// then combine their diagnostics.
+		pkg, _, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.TypecheckFull, source.NarrowestPackage)
 		if err != nil {
 			return nil, err
 		}
-		analysisDiags, err := source.Analyze(ctx, snapshot, pkg, true)
+		pkgDiags, err := pkg.DiagnosticsForFile(ctx, snapshot, uri)
 		if err != nil {
 			return nil, err
 		}
-		fileDiags := append(pkgDiagnostics[uri], analysisDiags[uri]...)
+		analysisDiags, err := source.Analyze(ctx, snapshot, pkg.Metadata().ID, true)
+		if err != nil {
+			return nil, err
+		}
+		var fileDiags []*source.Diagnostic
+		source.CombineDiagnostics(pkgDiags, analysisDiags[uri], &fileDiags, &fileDiags)
 
 		// Split diagnostics into fixes, which must match incoming diagnostics,
 		// and non-fixes, which must match the requested range. Build actions
@@ -212,7 +215,7 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 		}
 
 		if wanted[protocol.RefactorExtract] {
-			fixes, err := extractionFixes(ctx, snapshot, pkg, uri, params.Range)
+			fixes, err := extractionFixes(ctx, snapshot, uri, params.Range)
 			if err != nil {
 				return nil, err
 			}
@@ -305,7 +308,7 @@ func importDiagnostics(fix *imports.ImportFix, diagnostics []protocol.Diagnostic
 	return results
 }
 
-func extractionFixes(ctx context.Context, snapshot source.Snapshot, pkg source.Package, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
+func extractionFixes(ctx context.Context, snapshot source.Snapshot, uri span.URI, rng protocol.Range) ([]protocol.CodeAction, error) {
 	if rng.Start == rng.End {
 		return nil, nil
 	}
@@ -313,17 +316,17 @@ func extractionFixes(ctx context.Context, snapshot source.Snapshot, pkg source.P
 	if err != nil {
 		return nil, err
 	}
-	_, pgf, err := source.GetParsedFile(ctx, snapshot, fh, source.NarrowestPackage)
+	pgf, err := snapshot.ParseGo(ctx, fh, source.ParseFull)
 	if err != nil {
 		return nil, fmt.Errorf("getting file for Identifier: %w", err)
 	}
-	srng, err := pgf.Mapper.RangeToSpanRange(rng)
+	start, end, err := pgf.RangePos(rng)
 	if err != nil {
 		return nil, err
 	}
 	puri := protocol.URIFromSpanURI(uri)
 	var commands []protocol.Command
-	if _, ok, methodOk, _ := source.CanExtractFunction(pgf.Tok, srng, pgf.Src, pgf.File); ok {
+	if _, ok, methodOk, _ := source.CanExtractFunction(pgf.Tok, start, end, pgf.Src, pgf.File); ok {
 		cmd, err := command.NewApplyFixCommand("Extract function", command.ApplyFixArgs{
 			URI:   puri,
 			Fix:   source.ExtractFunction,
@@ -345,7 +348,7 @@ func extractionFixes(ctx context.Context, snapshot source.Snapshot, pkg source.P
 			commands = append(commands, cmd)
 		}
 	}
-	if _, _, ok, _ := source.CanExtractVariable(srng, pgf.File); ok {
+	if _, _, ok, _ := source.CanExtractVariable(start, end, pgf.File); ok {
 		cmd, err := command.NewApplyFixCommand("Extract variable", command.ApplyFixArgs{
 			URI:   puri,
 			Fix:   source.ExtractVariable,
@@ -367,7 +370,7 @@ func extractionFixes(ctx context.Context, snapshot source.Snapshot, pkg source.P
 	return actions, nil
 }
 
-func documentChanges(fh source.VersionedFileHandle, edits []protocol.TextEdit) []protocol.DocumentChanges {
+func documentChanges(fh source.FileHandle, edits []protocol.TextEdit) []protocol.DocumentChanges {
 	return []protocol.DocumentChanges{
 		{
 			TextDocumentEdit: &protocol.TextDocumentEdit{
@@ -411,7 +414,7 @@ func codeActionsForDiagnostic(ctx context.Context, snapshot source.Snapshot, sd
 	for _, fix := range sd.SuggestedFixes {
 		var changes []protocol.DocumentChanges
 		for uri, edits := range fix.Edits {
-			fh, err := snapshot.GetVersionedFile(ctx, uri)
+			fh, err := snapshot.GetFile(ctx, uri)
 			if err != nil {
 				return nil, err
 			}
