@@ -76,10 +76,11 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 		// this is a little cumbersome to avoid both exporting 'encoded' and its methods
 		// and to avoid import cycles
 		e := &encoded{
-			ctx:      ctx,
-			rng:      rng,
-			tokTypes: s.session.Options().SemanticTypes,
-			tokMods:  s.session.Options().SemanticMods,
+			ctx:            ctx,
+			metadataSource: snapshot,
+			rng:            rng,
+			tokTypes:       s.session.Options().SemanticTypes,
+			tokMods:        s.session.Options().SemanticMods,
 		}
 		add := func(line, start uint32, len uint32) {
 			e.add(line, start, len, tokMacro, nil)
@@ -92,11 +93,7 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 	if kind != source.Go {
 		return nil, nil
 	}
-	pkg, err := snapshot.PackageForFile(ctx, fh.URI(), source.TypecheckFull, source.WidestPackage)
-	if err != nil {
-		return nil, err
-	}
-	pgf, err := pkg.File(fh.URI())
+	pkg, pgf, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.TypecheckFull, source.NarrowestPackage)
 	if err != nil {
 		return nil, err
 	}
@@ -107,16 +104,17 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 		return nil, err
 	}
 	e := &encoded{
-		ctx:       ctx,
-		pgf:       pgf,
-		rng:       rng,
-		ti:        pkg.GetTypesInfo(),
-		pkg:       pkg,
-		fset:      pkg.FileSet(),
-		tokTypes:  s.session.Options().SemanticTypes,
-		tokMods:   s.session.Options().SemanticMods,
-		noStrings: vv.Options().NoSemanticString,
-		noNumbers: vv.Options().NoSemanticNumber,
+		ctx:            ctx,
+		metadataSource: snapshot,
+		pgf:            pgf,
+		rng:            rng,
+		ti:             pkg.GetTypesInfo(),
+		pkg:            pkg,
+		fset:           pkg.FileSet(),
+		tokTypes:       s.session.Options().SemanticTypes,
+		tokMods:        s.session.Options().SemanticMods,
+		noStrings:      vv.Options().NoSemanticString,
+		noNumbers:      vv.Options().NoSemanticNumber,
 	}
 	if err := e.init(); err != nil {
 		// e.init should never return an error, unless there's some
@@ -178,7 +176,6 @@ const (
 )
 
 func (e *encoded) token(start token.Pos, leng int, typ tokenType, mods []string) {
-
 	if !start.IsValid() {
 		// This is not worth reporting. TODO(pjw): does it still happen?
 		return
@@ -186,17 +183,10 @@ func (e *encoded) token(start token.Pos, leng int, typ tokenType, mods []string)
 	if start >= e.end || start+token.Pos(leng) <= e.start {
 		return
 	}
-	// want a line and column from start (in LSP coordinates)
-	// [//line directives should be ignored]
-	rng := source.NewMappedRange(e.pgf.Tok, e.pgf.Mapper, start, start+token.Pos(leng))
-	lspRange, err := rng.Range()
+	// want a line and column from start (in LSP coordinates). Ignore line directives.
+	lspRange, err := e.pgf.PosRange(start, start+token.Pos(leng))
 	if err != nil {
-		// possibly a //line directive. TODO(pjw): fix this somehow
-		// "column mapper is for file...instead of..."
-		// "line is beyond end of file..."
-		// see line 116 of internal/span/token.go which uses Position not PositionFor
-		// (it is too verbose to print the error on every token. some other RPC will fail)
-		// event.Error(e.ctx, "failed to convert to range", err)
+		event.Error(e.ctx, "failed to convert to range", err)
 		return
 	}
 	if lspRange.End.Line != lspRange.Start.Line {
@@ -228,7 +218,9 @@ type encoded struct {
 	noStrings bool
 	noNumbers bool
 
-	ctx               context.Context
+	ctx context.Context
+	// metadataSource is used to resolve imports
+	metadataSource    source.MetadataSource
 	tokTypes, tokMods []string
 	pgf               *source.ParsedGoFile
 	rng               *protocol.Range
@@ -251,14 +243,12 @@ func (e *encoded) strStack() string {
 	}
 	if len(e.stack) > 0 {
 		loc := e.stack[len(e.stack)-1].Pos()
-		if !safetoken.InRange(e.pgf.Tok, loc) {
+		if _, err := safetoken.Offset(e.pgf.Tok, loc); err != nil {
 			msg = append(msg, fmt.Sprintf("invalid position %v for %s", loc, e.pgf.URI))
-		} else if safetoken.InRange(e.pgf.Tok, loc) {
-			add := e.pgf.Tok.PositionFor(loc, false)
+		} else {
+			add := safetoken.Position(e.pgf.Tok, loc)
 			nm := filepath.Base(add.Filename)
 			msg = append(msg, fmt.Sprintf("(%s:%d,col:%d)", nm, add.Line, add.Column))
-		} else {
-			msg = append(msg, fmt.Sprintf("(loc %d out of range)", loc))
 		}
 	}
 	msg = append(msg, "]")
@@ -426,7 +416,7 @@ func (e *encoded) inspector(n ast.Node) bool {
 		return true
 	// not going to see these
 	case *ast.File, *ast.Package:
-		e.unexpected(fmt.Sprintf("implement %T %s", x, e.pgf.Tok.PositionFor(x.Pos(), false)))
+		e.unexpected(fmt.Sprintf("implement %T %s", x, safetoken.Position(e.pgf.Tok, x.Pos())))
 	// other things we knowingly ignore
 	case *ast.Comment, *ast.CommentGroup:
 		pop()
@@ -751,7 +741,11 @@ func (e *encoded) definitionFor(x *ast.Ident, def types.Object) (tokenType, []st
 				return tokFunction, mods
 			}
 			// if x < ... < FieldList < FuncDecl, this is the receiver, a variable
+			// PJW: maybe not. it might be a typeparameter in the type of the receiver
 			if _, ok := e.stack[i+1].(*ast.FieldList); ok {
+				if _, ok := def.(*types.TypeName); ok {
+					return tokTypeParam, mods
+				}
 				return tokVariable, nil
 			}
 			// if x < ... < FieldList < FuncType < FuncDecl, this is a param
@@ -786,7 +780,7 @@ func (e *encoded) definitionFor(x *ast.Ident, def types.Object) (tokenType, []st
 		}
 	}
 	// can't happen
-	msg := fmt.Sprintf("failed to find the decl for %s", e.pgf.Tok.PositionFor(x.Pos(), false))
+	msg := fmt.Sprintf("failed to find the decl for %s", safetoken.Position(e.pgf.Tok, x.Pos()))
 	e.unexpected(msg)
 	return "", []string{""}
 }
@@ -816,8 +810,8 @@ func (e *encoded) multiline(start, end token.Pos, val string, tok tokenType) {
 		}
 		return int(f.LineStart(line+1) - n)
 	}
-	spos := e.fset.PositionFor(start, false)
-	epos := e.fset.PositionFor(end, false)
+	spos := safetoken.StartPosition(e.fset, start)
+	epos := safetoken.EndPosition(e.fset, end)
 	sline := spos.Line
 	eline := epos.Line
 	// first line is from spos.Column to end
@@ -840,7 +834,7 @@ func (e *encoded) findKeyword(keyword string, start, end token.Pos) token.Pos {
 		return start + token.Pos(idx)
 	}
 	//(in unparsable programs: type _ <-<-chan int)
-	e.unexpected(fmt.Sprintf("not found:%s %v", keyword, e.fset.PositionFor(start, false)))
+	e.unexpected(fmt.Sprintf("not found:%s %v", keyword, safetoken.StartPosition(e.fset, start)))
 	return token.NoPos
 }
 
@@ -922,20 +916,24 @@ func (e *encoded) importSpec(d *ast.ImportSpec) {
 		return
 	}
 	// Import strings are implementation defined. Try to match with parse information.
-	imported, err := e.pkg.ResolveImportPath(importPath)
-	if err != nil {
+	depID := e.pkg.Metadata().DepsByImpPath[importPath]
+	if depID == "" {
+		return
+	}
+	depMD := e.metadataSource.Metadata(depID)
+	if depMD == nil {
 		// unexpected, but impact is that maybe some import is not colored
 		return
 	}
 	// Check whether the original literal contains the package's declared name.
-	j := strings.LastIndex(d.Path.Value, string(imported.Name()))
+	j := strings.LastIndex(d.Path.Value, string(depMD.Name))
 	if j == -1 {
-		// name doesn't show up, for whatever reason, so nothing to report
+		// Package name does not match import path, so there is nothing to report.
 		return
 	}
 	// Report virtual declaration at the position of the substring.
 	start := d.Path.Pos() + token.Pos(j)
-	e.token(start, len(imported.Name()), tokNamespace, nil)
+	e.token(start, len(depMD.Name), tokNamespace, nil)
 }
 
 // log unexpected state
