@@ -26,7 +26,6 @@ import (
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/snippet"
 	"golang.org/x/tools/gopls/internal/lsp/source"
-	"golang.org/x/tools/gopls/internal/span"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/fuzzy"
 	"golang.org/x/tools/internal/imports"
@@ -157,7 +156,8 @@ func (ipm insensitivePrefixMatcher) Score(candidateLabel string) float32 {
 type completer struct {
 	snapshot source.Snapshot
 	pkg      source.Package
-	qf       types.Qualifier
+	qf       types.Qualifier          // for qualifying typed expressions
+	mq       source.MetadataQualifier // for syntactic qualifying
 	opts     *completionOptions
 
 	// completionContext contains information about the trigger for this
@@ -218,7 +218,7 @@ type completer struct {
 	methodSetCache map[methodSetKey]*types.MethodSet
 
 	// mapper converts the positions in the file from which the completion originated.
-	mapper *protocol.ColumnMapper
+	mapper *protocol.Mapper
 
 	// startTime is when we started processing this completion request. It does
 	// not include any time the request spent in the queue.
@@ -288,25 +288,26 @@ type completionContext struct {
 
 // A Selection represents the cursor position and surrounding identifier.
 type Selection struct {
-	content string
-	cursor  token.Pos // relative to rng.TokFile
-	rng     span.Range
+	content            string
+	tokFile            *token.File
+	start, end, cursor token.Pos // relative to rng.TokFile
+	mapper             *protocol.Mapper
 }
 
 func (p Selection) Content() string {
 	return p.content
 }
 
-func (p Selection) Range() span.Range {
-	return p.rng
+func (p Selection) Range() (protocol.Range, error) {
+	return p.mapper.PosRange(p.tokFile, p.start, p.end)
 }
 
 func (p Selection) Prefix() string {
-	return p.content[:p.cursor-p.rng.Start]
+	return p.content[:p.cursor-p.start]
 }
 
 func (p Selection) Suffix() string {
-	return p.content[p.cursor-p.rng.Start:]
+	return p.content[p.cursor-p.start:]
 }
 
 func (c *completer) setSurrounding(ident *ast.Ident) {
@@ -321,7 +322,10 @@ func (c *completer) setSurrounding(ident *ast.Ident) {
 		content: ident.Name,
 		cursor:  c.pos,
 		// Overwrite the prefix only.
-		rng: span.NewRange(c.tokFile, ident.Pos(), ident.End()),
+		tokFile: c.tokFile,
+		start:   ident.Pos(),
+		end:     ident.End(),
+		mapper:  c.mapper,
 	}
 
 	c.setMatcherFromPrefix(c.surrounding.Prefix())
@@ -343,7 +347,10 @@ func (c *completer) getSurrounding() *Selection {
 		c.surrounding = &Selection{
 			content: "",
 			cursor:  c.pos,
-			rng:     span.NewRange(c.tokFile, c.pos, c.pos),
+			tokFile: c.tokFile,
+			start:   c.pos,
+			end:     c.pos,
+			mapper:  c.mapper,
 		}
 	}
 	return c.surrounding
@@ -427,7 +434,7 @@ func Completion(ctx context.Context, snapshot source.Snapshot, fh source.FileHan
 
 	startTime := time.Now()
 
-	pkg, pgf, err := source.GetParsedFile(ctx, snapshot, fh, source.NarrowestPackage)
+	pkg, pgf, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.TypecheckFull, source.NarrowestPackage)
 	if err != nil || pgf.File.Package == token.NoPos {
 		// If we can't parse this file or find position for the package
 		// keyword, it may be missing a package declaration. Try offering
@@ -441,7 +448,7 @@ func Completion(ctx context.Context, snapshot source.Snapshot, fh source.FileHan
 		}
 		return items, surrounding, nil
 	}
-	pos, err := pgf.Mapper.Pos(protoPos)
+	pos, err := pgf.PositionPos(protoPos)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -482,7 +489,7 @@ func Completion(ctx context.Context, snapshot source.Snapshot, fh source.FileHan
 					qual := types.RelativeTo(pkg.GetTypes())
 					objStr = types.ObjectString(obj, qual)
 				}
-				ans, sel := definition(path, obj, pgf.Tok, fh)
+				ans, sel := definition(path, obj, pgf)
 				if ans != nil {
 					sort.Slice(ans, func(i, j int) bool {
 						return ans[i].Score > ans[j].Score
@@ -503,6 +510,7 @@ func Completion(ctx context.Context, snapshot source.Snapshot, fh source.FileHan
 		pkg:      pkg,
 		snapshot: snapshot,
 		qf:       source.Qualifier(pgf.File, pkg.GetTypes(), pkg.GetTypesInfo()),
+		mq:       source.MetadataQualifierForFile(snapshot, pgf.File, pkg.Metadata()),
 		completionContext: completionContext{
 			triggerCharacter: protoContext.TriggerCharacter,
 			triggerKind:      protoContext.TriggerKind,
@@ -795,7 +803,10 @@ func (c *completer) populateImportCompletions(ctx context.Context, searchImport
 	c.surrounding = &Selection{
 		content: content,
 		cursor:  c.pos,
-		rng:     span.NewRange(c.tokFile, start, end),
+		tokFile: c.tokFile,
+		start:   start,
+		end:     end,
+		mapper:  c.mapper,
 	}
 
 	seenImports := make(map[string]struct{})
@@ -1015,7 +1026,10 @@ func (c *completer) setSurroundingForComment(comments *ast.CommentGroup) {
 	c.surrounding = &Selection{
 		content: cursorComment.Text[start:end],
 		cursor:  c.pos,
-		rng:     span.NewRange(c.tokFile, token.Pos(int(cursorComment.Slash)+start), token.Pos(int(cursorComment.Slash)+end)),
+		tokFile: c.tokFile,
+		start:   token.Pos(int(cursorComment.Slash) + start),
+		end:     token.Pos(int(cursorComment.Slash) + end),
+		mapper:  c.mapper,
 	}
 	c.setMatcherFromPrefix(c.surrounding.Prefix())
 }
@@ -1083,10 +1097,12 @@ func (c *completer) selector(ctx context.Context, sel *ast.SelectorExpr) error {
 	// Is sel a qualified identifier?
 	if id, ok := sel.X.(*ast.Ident); ok {
 		if pkgName, ok := c.pkg.GetTypesInfo().Uses[id].(*types.PkgName); ok {
-			pkg, _ := c.pkg.DirectDep(source.PackagePath(pkgName.Imported().Path()))
-			// If the package is not imported, try searching for unimported
-			// completions.
-			if pkg == nil && c.opts.unimported {
+			// If this package path is not a known dep, it means that it resolves to
+			// an import path that couldn't be resolved by go/packages.
+			//
+			// Try to complete from the package cache.
+			pkgPath := source.PackagePath(pkgName.Imported().Path())
+			if _, ok := c.pkg.Metadata().DepsByPkgPath[pkgPath]; !ok && c.opts.unimported {
 				if err := c.unimportedMembers(ctx, id); err != nil {
 					return err
 				}
@@ -1128,7 +1144,7 @@ func (c *completer) unimportedMembers(ctx context.Context, id *ast.Ident) error
 
 	var paths []string
 	for path, pkg := range known {
-		if pkg.GetTypes().Name() != id.Name {
+		if pkg.Name() != id.Name {
 			continue
 		}
 		paths = append(paths, string(path))
@@ -1150,16 +1166,16 @@ func (c *completer) unimportedMembers(ctx context.Context, id *ast.Ident) error
 
 	for _, path := range paths {
 		pkg := known[source.PackagePath(path)]
-		if pkg.GetTypes().Name() != id.Name {
+		if pkg.Name() != id.Name {
 			continue
 		}
 		imp := &importInfo{
 			importPath: path,
 		}
-		if imports.ImportPathToAssumedName(path) != pkg.GetTypes().Name() {
-			imp.name = pkg.GetTypes().Name()
+		if imports.ImportPathToAssumedName(path) != pkg.Name() {
+			imp.name = pkg.Name()
 		}
-		c.packageMembers(pkg.GetTypes(), unimportedScore(relevances[path]), imp, func(cand candidate) {
+		c.packageMembers(pkg, unimportedScore(relevances[path]), imp, func(cand candidate) {
 			c.deepState.enqueue(cand)
 		})
 		if len(c.items) >= unimportedMemberTarget {
@@ -1282,12 +1298,10 @@ func (c *completer) lexical(ctx context.Context) error {
 	var (
 		builtinIota = types.Universe.Lookup("iota")
 		builtinNil  = types.Universe.Lookup("nil")
-		// comparable is an interface that exists on the dev.typeparams Go branch.
-		// Filter it out from completion results to stabilize tests.
-		// TODO(rFindley) update (or remove) our handling for comparable once the
-		//                type parameter API has stabilized.
-		builtinAny        = types.Universe.Lookup("any")
-		builtinComparable = types.Universe.Lookup("comparable")
+
+		// TODO(rfindley): only allow "comparable" where it is valid (in constraint
+		// position or embedded in interface declarations).
+		// builtinComparable = types.Universe.Lookup("comparable")
 	)
 
 	// Track seen variables to avoid showing completions for shadowed variables.
@@ -1306,9 +1320,6 @@ func (c *completer) lexical(ctx context.Context) error {
 			if declScope != scope {
 				continue // Name was declared in some enclosing scope, or not at all.
 			}
-			if obj == builtinComparable || obj == builtinAny {
-				continue
-			}
 
 			// If obj's type is invalid, find the AST node that defines the lexical block
 			// containing the declaration of obj. Don't resolve types for packages.
@@ -1417,7 +1428,7 @@ func (c *completer) lexical(ctx context.Context) error {
 	return nil
 }
 
-// injectType manufacters candidates based on the given type. This is
+// injectType manufactures candidates based on the given type. This is
 // intended for types not discoverable via lexical search, such as
 // composite and/or generic types. For example, if the type is "[]int",
 // this method makes sure you get candidates "[]int{}" and "[]int"
@@ -1479,7 +1490,7 @@ func (c *completer) unimportedPackages(ctx context.Context, seen map[string]stru
 	}
 	var paths []string // actually PackagePaths
 	for path, pkg := range known {
-		if !strings.HasPrefix(pkg.GetTypes().Name(), prefix) {
+		if !strings.HasPrefix(pkg.Name(), prefix) {
 			continue
 		}
 		paths = append(paths, string(path))
@@ -1508,21 +1519,21 @@ func (c *completer) unimportedPackages(ctx context.Context, seen map[string]stru
 
 	for _, path := range paths {
 		pkg := known[source.PackagePath(path)]
-		if _, ok := seen[pkg.GetTypes().Name()]; ok {
+		if _, ok := seen[pkg.Name()]; ok {
 			continue
 		}
 		imp := &importInfo{
 			importPath: path,
 		}
-		if imports.ImportPathToAssumedName(path) != pkg.GetTypes().Name() {
-			imp.name = pkg.GetTypes().Name()
+		if imports.ImportPathToAssumedName(path) != pkg.Name() {
+			imp.name = pkg.Name()
 		}
 		if count >= maxUnimportedPackageNames {
 			return nil
 		}
 		c.deepState.enqueue(candidate{
 			// Pass an empty *types.Package to disable deep completions.
-			obj:   types.NewPkgName(0, nil, pkg.GetTypes().Name(), types.NewPackage(path, string(pkg.Name()))),
+			obj:   types.NewPkgName(0, nil, pkg.Name(), types.NewPackage(path, string(pkg.Name()))),
 			score: unimportedScore(relevances[path]),
 			imp:   imp,
 		})
@@ -2917,7 +2928,7 @@ func (c *completer) matchingTypeName(cand *candidate) bool {
 		return true
 	}
 
-	if !source.IsInterface(t) && typeMatches(types.NewPointer(t)) {
+	if !types.IsInterface(t) && typeMatches(types.NewPointer(t)) {
 		if c.inference.typeName.compLitType {
 			// If we are completing a composite literal type as in
 			// "foo<>{}", to make a pointer we must prepend "&".
