@@ -15,14 +15,15 @@ import (
 	"go/token"
 	"go/types"
 	"io"
-	"strings"
 
 	"golang.org/x/mod/modfile"
-	"golang.org/x/mod/module"
 	"golang.org/x/tools/go/analysis"
 	"golang.org/x/tools/go/packages"
+	"golang.org/x/tools/go/types/objectpath"
 	"golang.org/x/tools/gopls/internal/govulncheck"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
+	"golang.org/x/tools/gopls/internal/lsp/safetoken"
+	"golang.org/x/tools/gopls/internal/lsp/source/methodsets"
 	"golang.org/x/tools/gopls/internal/span"
 	"golang.org/x/tools/internal/event/label"
 	"golang.org/x/tools/internal/event/tag"
@@ -74,16 +75,9 @@ type Snapshot interface {
 	// and their GOPATH.
 	ValidBuildConfiguration() bool
 
-	// WriteEnv writes the view-specific environment to the io.Writer.
-	WriteEnv(ctx context.Context, w io.Writer) error
-
 	// FindFile returns the FileHandle for the given URI, if it is already
 	// in the given snapshot.
-	FindFile(uri span.URI) VersionedFileHandle
-
-	// GetVersionedFile returns the VersionedFileHandle for a given URI,
-	// initializing it if it is not already part of the snapshot.
-	GetVersionedFile(ctx context.Context, uri span.URI) (VersionedFileHandle, error)
+	FindFile(uri span.URI) FileHandle
 
 	// GetFile returns the FileHandle for a given URI, initializing it if it is
 	// not already part of the snapshot.
@@ -100,18 +94,15 @@ type Snapshot interface {
 	IgnoredFile(uri span.URI) bool
 
 	// Templates returns the .tmpl files
-	Templates() map[span.URI]VersionedFileHandle
+	Templates() map[span.URI]FileHandle
 
 	// ParseGo returns the parsed AST for the file.
 	// If the file is not available, returns nil and an error.
+	// Position information is added to FileSet().
 	ParseGo(ctx context.Context, fh FileHandle, mode ParseMode) (*ParsedGoFile, error)
 
-	// DiagnosePackage returns basic diagnostics, including list, parse, and type errors
-	// for pkg, grouped by file.
-	DiagnosePackage(ctx context.Context, pkg Package) (map[span.URI][]*Diagnostic, error)
-
-	// Analyze runs the analyses for the given package at this snapshot.
-	Analyze(ctx context.Context, pkgID PackageID, analyzers []*Analyzer) ([]*Diagnostic, error)
+	// Analyze runs the specified analyzers on the given package at this snapshot.
+	Analyze(ctx context.Context, id PackageID, analyzers []*Analyzer) ([]*Diagnostic, error)
 
 	// RunGoCommandPiped runs the given `go` command, writing its output
 	// to stdout and stderr. Verb, Args, and WorkingDir must be specified.
@@ -167,60 +158,53 @@ type Snapshot interface {
 	// IsBuiltin reports whether uri is part of the builtin package.
 	IsBuiltin(ctx context.Context, uri span.URI) bool
 
-	// PackagesForFile returns an unordered list of packages that contain
-	// the file denoted by uri, type checked in the specified mode.
-	//
-	// If withIntermediateTestVariants is set, the resulting package set includes
-	// intermediate test variants.
-	PackagesForFile(ctx context.Context, uri span.URI, mode TypecheckMode, withIntermediateTestVariants bool) ([]Package, error)
-
-	// PackageForFile returns a single package that this file belongs to,
-	// checked in mode and filtered by the package policy.
-	PackageForFile(ctx context.Context, uri span.URI, mode TypecheckMode, selectPackage PackageFilter) (Package, error)
-
-	// GetActiveReverseDeps returns the active files belonging to the reverse
-	// dependencies of this file's package, checked in TypecheckWorkspace mode.
-	GetReverseDependencies(ctx context.Context, id PackageID) ([]Package, error)
+	// ReverseDependencies returns a new mapping whose entries are
+	// the ID and Metadata of each package in the workspace that
+	// directly or transitively depend on the package denoted by id,
+	// excluding id itself.
+	ReverseDependencies(ctx context.Context, id PackageID, transitive bool) (map[PackageID]*Metadata, error)
 
 	// CachedImportPaths returns all the imported packages loaded in this
 	// snapshot, indexed by their package path (not import path, despite the name)
 	// and checked in TypecheckWorkspace mode.
-	CachedImportPaths(ctx context.Context) (map[PackagePath]Package, error)
-
-	// KnownPackages returns a new unordered list of all packages
-	// loaded in this snapshot, checked in TypecheckWorkspace mode.
 	//
-	// TODO(adonovan): opt: rewrite 'implementations' to avoid the
-	// need ever to "load everything at once" using this function.
-	KnownPackages(ctx context.Context) ([]Package, error)
+	// To reduce latency, it does not wait for type-checking to complete.
+	// It is intended for use only in completions.
+	CachedImportPaths(ctx context.Context) (map[PackagePath]*types.Package, error)
 
-	// ActivePackages returns the packages considered 'active' in the workspace.
+	// ActiveMetadata returns a new, unordered slice containing
+	// metadata for all packages considered 'active' in the workspace.
 	//
 	// In normal memory mode, this is all workspace packages. In degraded memory
 	// mode, this is just the reverse transitive closure of open packages.
-	ActivePackages(ctx context.Context) ([]Package, error)
+	ActiveMetadata(ctx context.Context) ([]*Metadata, error)
 
-	// AllValidMetadata returns all valid metadata loaded for the snapshot.
-	AllValidMetadata(ctx context.Context) ([]*Metadata, error)
-
-	// WorkspacePackageByID returns the workspace package with id, type checked
-	// in 'workspace' mode.
-	WorkspacePackageByID(ctx context.Context, id PackageID) (Package, error)
+	// AllMetadata returns a new unordered array of metadata for all packages in the workspace.
+	AllMetadata(ctx context.Context) ([]*Metadata, error)
 
 	// Symbols returns all symbols in the snapshot.
 	Symbols(ctx context.Context) map[span.URI][]Symbol
 
-	// Metadata returns metadata for each package associated with the given file URI.
+	// Metadata returns the metadata for the specified package,
+	// or nil if it was not found.
+	Metadata(id PackageID) *Metadata
+
+	// MetadataForFile returns a new slice containing metadata for each
+	// package containing the Go file identified by uri, ordered by the
+	// number of CompiledGoFiles (i.e. "narrowest" to "widest" package).
+	// The result may include tests and intermediate test variants of
+	// importable packages.
+	// It returns an error if the context was cancelled.
 	MetadataForFile(ctx context.Context, uri span.URI) ([]*Metadata, error)
 
+	// TypeCheck parses and type-checks the specified packages,
+	// and returns them in the same order as the ids.
+	TypeCheck(ctx context.Context, mode TypecheckMode, ids ...PackageID) ([]Package, error)
+
 	// GetCriticalError returns any critical errors in the workspace.
 	//
 	// A nil result may mean success, or context cancellation.
 	GetCriticalError(ctx context.Context) *CriticalError
-
-	// BuildGoplsMod generates a go.mod file for all modules in the workspace.
-	// It bypasses any existing gopls.mod.
-	BuildGoplsMod(ctx context.Context) (*modfile.File, error)
 }
 
 // SnapshotLabels returns a new slice of labels that should be used for events
@@ -229,16 +213,49 @@ func SnapshotLabels(snapshot Snapshot) []label.Label {
 	return []label.Label{tag.Snapshot.Of(snapshot.SequenceID()), tag.Directory.Of(snapshot.View().Folder())}
 }
 
-// PackageFilter sets how a package is filtered out from a set of packages
+// PackageForFile is a convenience function that selects a package to
+// which this file belongs (narrowest or widest), type-checks it in
+// the requested mode (full or workspace), and returns it, along with
+// the parse tree of that file.
+//
+// Type-checking is expensive. Call snapshot.ParseGo if all you need
+// is a parse tree, or snapshot.MetadataForFile if you only need metadata.
+func PackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI, mode TypecheckMode, pkgSel PackageSelector) (Package, *ParsedGoFile, error) {
+	metas, err := snapshot.MetadataForFile(ctx, uri)
+	if err != nil {
+		return nil, nil, err
+	}
+	if len(metas) == 0 {
+		return nil, nil, fmt.Errorf("no package metadata for file %s", uri)
+	}
+	switch pkgSel {
+	case NarrowestPackage:
+		metas = metas[:1]
+	case WidestPackage:
+		metas = metas[len(metas)-1:]
+	}
+	pkgs, err := snapshot.TypeCheck(ctx, mode, metas[0].ID)
+	if err != nil {
+		return nil, nil, err
+	}
+	pkg := pkgs[0]
+	pgf, err := pkg.File(uri)
+	if err != nil {
+		return nil, nil, err // "can't happen"
+	}
+	return pkg, pgf, err
+}
+
+// PackageSelector sets how a package is selected out from a set of packages
 // containing a given file.
-type PackageFilter int
+type PackageSelector int
 
 const (
 	// NarrowestPackage picks the "narrowest" package for a given file.
 	// By "narrowest" package, we mean the package with the fewest number of
 	// files that includes the given file. This solves the problem of test
 	// variants, as the test will have more files than the non-test package.
-	NarrowestPackage PackageFilter = iota
+	NarrowestPackage PackageSelector = iota
 
 	// WidestPackage returns the Package containing the most files.
 	// This is useful for something like diagnostics, where we'd prefer to
@@ -291,7 +308,10 @@ type View interface {
 	// Snapshot returns the current snapshot for the view, and a
 	// release function that must be called when the Snapshot is
 	// no longer needed.
-	Snapshot(ctx context.Context) (Snapshot, func())
+	//
+	// If the view is shut down, the resulting error will be non-nil, and the
+	// release function need not be called.
+	Snapshot() (Snapshot, func(), error)
 
 	// IsGoPrivatePath reports whether target is a private import path, as identified
 	// by the GOPRIVATE environment variable.
@@ -308,16 +328,22 @@ type View interface {
 	// ClearModuleUpgrades clears all upgrades for the modules in modfile.
 	ClearModuleUpgrades(modfile span.URI)
 
-	// Vulnerabilites returns known vulnerabilities for the given modfile.
+	// Vulnerabilities returns known vulnerabilities for the given modfile.
 	// TODO(suzmue): replace command.Vuln with a different type, maybe
 	// https://pkg.go.dev/golang.org/x/vuln/cmd/govulncheck/govulnchecklib#Summary?
 	Vulnerabilities(modfile ...span.URI) map[span.URI]*govulncheck.Result
 
-	// SetVulnerabilities resets the list of vulnerabilites that exists for the given modules
+	// SetVulnerabilities resets the list of vulnerabilities that exists for the given modules
 	// required by modfile.
 	SetVulnerabilities(modfile span.URI, vulncheckResult *govulncheck.Result)
 
-	// FileKind returns the type of a file
+	// FileKind returns the type of a file.
+	//
+	// We can't reliably deduce the kind from the file name alone,
+	// as some editors can be told to interpret a buffer as
+	// language different from the file name heuristic, e.g. that
+	// an .html file actually contains Go "html/template" syntax,
+	// or even that a .go file contains Python.
 	FileKind(FileHandle) FileKind
 
 	// GoVersion returns the configured Go version for this view.
@@ -328,14 +354,22 @@ type View interface {
 	GoVersionString() string
 }
 
-// A FileSource maps uris to FileHandles. This abstraction exists both for
-// testability, and so that algorithms can be run equally on session and
-// snapshot files.
+// A FileSource maps uris to FileHandles.
 type FileSource interface {
 	// GetFile returns the FileHandle for a given URI.
 	GetFile(ctx context.Context, uri span.URI) (FileHandle, error)
 }
 
+// A MetadataSource maps package IDs to metadata.
+//
+// TODO(rfindley): replace this with a concrete metadata graph, once it is
+// exposed from the snapshot.
+type MetadataSource interface {
+	// Metadata returns Metadata for the given package ID, or nil if it does not
+	// exist.
+	Metadata(PackageID) *Metadata
+}
+
 // A ParsedGoFile contains the results of parsing a Go file.
 type ParsedGoFile struct {
 	URI  span.URI
@@ -346,15 +380,67 @@ type ParsedGoFile struct {
 	// actual content of the file if we have fixed the AST.
 	Src      []byte
 	Fixed    bool
-	Mapper   *protocol.ColumnMapper
+	Mapper   *protocol.Mapper // may map fixed Src, not file content
 	ParseErr scanner.ErrorList
 }
 
+// -- go/token domain convenience helpers --
+
+// PositionPos returns the token.Pos of protocol position p within the file.
+func (pgf *ParsedGoFile) PositionPos(p protocol.Position) (token.Pos, error) {
+	offset, err := pgf.Mapper.PositionOffset(p)
+	if err != nil {
+		return token.NoPos, err
+	}
+	return safetoken.Pos(pgf.Tok, offset)
+}
+
+// PosRange returns a protocol Range for the token.Pos interval in this file.
+func (pgf *ParsedGoFile) PosRange(start, end token.Pos) (protocol.Range, error) {
+	return pgf.Mapper.PosRange(pgf.Tok, start, end)
+}
+
+// PosMappedRange returns a MappedRange for the token.Pos interval in this file.
+// A MappedRange can be converted to any other form.
+func (pgf *ParsedGoFile) PosMappedRange(start, end token.Pos) (protocol.MappedRange, error) {
+	return pgf.Mapper.PosMappedRange(pgf.Tok, start, end)
+}
+
+// PosLocation returns a protocol Location for the token.Pos interval in this file.
+func (pgf *ParsedGoFile) PosLocation(start, end token.Pos) (protocol.Location, error) {
+	return pgf.Mapper.PosLocation(pgf.Tok, start, end)
+}
+
+// NodeRange returns a protocol Range for the ast.Node interval in this file.
+func (pgf *ParsedGoFile) NodeRange(node ast.Node) (protocol.Range, error) {
+	return pgf.Mapper.NodeRange(pgf.Tok, node)
+}
+
+// NodeMappedRange returns a MappedRange for the ast.Node interval in this file.
+// A MappedRange can be converted to any other form.
+func (pgf *ParsedGoFile) NodeMappedRange(node ast.Node) (protocol.MappedRange, error) {
+	return pgf.Mapper.NodeMappedRange(pgf.Tok, node)
+}
+
+// NodeLocation returns a protocol Location for the ast.Node interval in this file.
+func (pgf *ParsedGoFile) NodeLocation(node ast.Node) (protocol.Location, error) {
+	return pgf.Mapper.PosLocation(pgf.Tok, node.Pos(), node.End())
+}
+
+// RangePos parses a protocol Range back into the go/token domain.
+func (pgf *ParsedGoFile) RangePos(r protocol.Range) (token.Pos, token.Pos, error) {
+	start, end, err := pgf.Mapper.RangeOffsets(r)
+	if err != nil {
+		return token.NoPos, token.NoPos, err
+	}
+	return pgf.Tok.Pos(start), pgf.Tok.Pos(end), nil
+}
+
 // A ParsedModule contains the results of parsing a go.mod file.
 type ParsedModule struct {
 	URI         span.URI
 	File        *modfile.File
-	Mapper      *protocol.ColumnMapper
+	Mapper      *protocol.Mapper
 	ParseErrors []*Diagnostic
 }
 
@@ -362,7 +448,7 @@ type ParsedModule struct {
 type ParsedWorkFile struct {
 	URI         span.URI
 	File        *modfile.WorkFile
-	Mapper      *protocol.ColumnMapper
+	Mapper      *protocol.Mapper
 	ParseErrors []*Diagnostic
 }
 
@@ -388,9 +474,7 @@ type Metadata struct {
 	DepsByPkgPath   map[PackagePath]PackageID // values are unique and non-empty
 	Module          *packages.Module
 	DepsErrors      []*packagesinternal.PackageError
-
-	// Config is the *packages.Config associated with the loaded package.
-	Config *packages.Config
+	LoadDir         string // directory from which go/packages was run
 }
 
 // IsIntermediateTestVariant reports whether the given package is an
@@ -427,14 +511,19 @@ func (m *Metadata) IsIntermediateTestVariant() bool {
 	return m.ForTest != "" && m.ForTest != m.PkgPath && m.ForTest+"_test" != m.PkgPath
 }
 
-var ErrViewExists = errors.New("view already exists for session")
-
-// Overlay is the type for a file held in memory on a session.
-type Overlay interface {
-	Kind() FileKind
-	VersionedFileHandle
+// RemoveIntermediateTestVariants removes intermediate test variants, modifying the array.
+func RemoveIntermediateTestVariants(metas []*Metadata) []*Metadata {
+	res := metas[:0]
+	for _, m := range metas {
+		if !m.IsIntermediateTestVariant() {
+			res = append(res, m)
+		}
+	}
+	return res
 }
 
+var ErrViewExists = errors.New("view already exists for session")
+
 // FileModification represents a modification to a file.
 type FileModification struct {
 	URI    span.URI
@@ -529,38 +618,26 @@ const (
 	TypecheckWorkspace
 )
 
-type VersionedFileHandle interface {
-	FileHandle
-	Version() int32
-	Session() string
-
-	// LSPIdentity returns the version identity of a file.
-	VersionedFileIdentity() VersionedFileIdentity
-}
-
-type VersionedFileIdentity struct {
-	URI span.URI
-
-	// SessionID is the ID of the LSP session.
-	SessionID string
-
-	// Version is the version of the file, as specified by the client. It should
-	// only be set in combination with SessionID.
-	Version int32
-}
-
-// FileHandle represents a handle to a specific version of a single file.
+// A FileHandle is an interface to files tracked by the LSP session, which may
+// be either files read from disk, or open in the editor session (overlays).
 type FileHandle interface {
+	// URI is the URI for this file handle.
+	// TODO(rfindley): this is not actually well-defined. In some cases, there
+	// may be more than one URI that resolve to the same FileHandle. Which one is
+	// this?
 	URI() span.URI
-
 	// FileIdentity returns a FileIdentity for the file, even if there was an
 	// error reading it.
 	FileIdentity() FileIdentity
+	// Saved reports whether the file has the same content on disk.
+	// For on-disk files, this is trivially true.
+	Saved() bool
+	// Version returns the file version, as defined by the LSP client.
+	// For on-disk file handles, Version returns 0.
+	Version() int32
 	// Read reads the contents of a file.
 	// If the file is not available, returns a nil slice and an error.
 	Read() ([]byte, error)
-	// Saved reports whether the file has the same content on disk.
-	Saved() bool
 }
 
 // A Hash is a cryptographic digest of the contents of a file.
@@ -628,6 +705,23 @@ const (
 	Work
 )
 
+func (k FileKind) String() string {
+	switch k {
+	case Go:
+		return "go"
+	case Mod:
+		return "go.mod"
+	case Sum:
+		return "go.sum"
+	case Tmpl:
+		return "tmpl"
+	case Work:
+		return "go.work"
+	default:
+		return fmt.Sprintf("internal error: unknown file kind %d", k)
+	}
+}
+
 // Analyzer represents a go/analysis analyzer with some boolean properties
 // that let the user know how to use the analyzer.
 type Analyzer struct {
@@ -636,6 +730,8 @@ type Analyzer struct {
 	// Enabled reports whether the analyzer is enabled. This value can be
 	// configured per-analysis in user settings. For staticcheck analyzers,
 	// the value of the Staticcheck setting overrides this field.
+	//
+	// Most clients should use the IsEnabled method.
 	Enabled bool
 
 	// Fix is the name of the suggested fix name used to invoke the suggested
@@ -653,14 +749,17 @@ type Analyzer struct {
 	Severity protocol.DiagnosticSeverity
 }
 
-func (a Analyzer) IsEnabled(view View) bool {
+func (a *Analyzer) String() string { return a.Analyzer.String() }
+
+// IsEnabled reports whether this analyzer is enabled by the given options.
+func (a Analyzer) IsEnabled(options *Options) bool {
 	// Staticcheck analyzers can only be enabled when staticcheck is on.
-	if _, ok := view.Options().StaticcheckAnalyzers[a.Analyzer.Name]; ok {
-		if !view.Options().Staticcheck {
+	if _, ok := options.StaticcheckAnalyzers[a.Analyzer.Name]; ok {
+		if !options.Staticcheck {
 			return false
 		}
 	}
-	if enabled, ok := view.Options().Analyses[a.Analyzer.Name]; ok {
+	if enabled, ok := options.Analyses[a.Analyzer.Name]; ok {
 		return enabled
 	}
 	return a.Enabled
@@ -680,13 +779,7 @@ type (
 // Package represents a Go package that has been parsed and type-checked.
 // It maintains only the relevant fields of a *go/packages.Package.
 type Package interface {
-	// Metadata:
-	ID() PackageID
-	Name() PackageName
-	PkgPath() PackagePath
-	GetTypesSizes() types.Sizes
-	ForTest() string
-	Version() *module.Version // may differ from Metadata.Module.Version
+	Metadata() *Metadata
 
 	// Results of parsing:
 	FileSet() *token.FileSet
@@ -694,17 +787,19 @@ type Package interface {
 	CompiledGoFiles() []*ParsedGoFile // (borrowed)
 	File(uri span.URI) (*ParsedGoFile, error)
 	GetSyntax() []*ast.File // (borrowed)
-	HasListOrParseErrors() bool
+	HasParseErrors() bool
 
 	// Results of type checking:
 	GetTypes() *types.Package
 	GetTypesInfo() *types.Info
-	DirectDep(path PackagePath) (Package, error)
-	ResolveImportPath(path ImportPath) (Package, error)
-	Imports() []Package // new slice of all direct dependencies, unordered
 	HasTypeErrors() bool
+	DiagnosticsForFile(ctx context.Context, s Snapshot, uri span.URI) ([]*Diagnostic, error)
+	ReferencesTo(map[PackagePath]map[objectpath.Path]unit) []protocol.Location
+	MethodSetsIndex() *methodsets.Index
 }
 
+type unit = struct{}
+
 // A CriticalError is a workspace-wide error that generally prevents gopls from
 // functioning correctly. In the presence of critical errors, other diagnostics
 // in the workspace may not make sense.
@@ -737,7 +832,6 @@ type Diagnostic struct {
 	// Fields below are used internally to generate quick fixes. They aren't
 	// part of the LSP spec and don't leave the server.
 	SuggestedFixes []SuggestedFix
-	Analyzer       *Analyzer
 }
 
 func (d *Diagnostic) String() string {
@@ -754,7 +848,8 @@ const (
 	ModTidyError             DiagnosticSource = "go mod tidy"
 	OptimizationDetailsError DiagnosticSource = "optimizer details"
 	UpgradeNotification      DiagnosticSource = "upgrade available"
-	Vulncheck                DiagnosticSource = "govulncheck"
+	Vulncheck                DiagnosticSource = "vulncheck imports"
+	Govulncheck              DiagnosticSource = "govulncheck"
 	TemplateError            DiagnosticSource = "template"
 	WorkFileError            DiagnosticSource = "go.work file"
 )
@@ -762,28 +857,3 @@ const (
 func AnalyzerErrorKind(name string) DiagnosticSource {
 	return DiagnosticSource(name)
 }
-
-// WorkspaceModuleVersion is the nonexistent pseudoversion suffix used in the
-// construction of the workspace module. It is exported so that we can make
-// sure not to show this version to end users in error messages, to avoid
-// confusion.
-// The major version is not included, as that depends on the module path.
-//
-// If workspace module A is dependent on workspace module B, we need our
-// nonexistent version to be greater than the version A mentions.
-// Otherwise, the go command will try to update to that version. Use a very
-// high minor version to make that more likely.
-const workspaceModuleVersion = ".9999999.0-goplsworkspace"
-
-func IsWorkspaceModuleVersion(version string) bool {
-	return strings.HasSuffix(version, workspaceModuleVersion)
-}
-
-func WorkspaceModuleVersion(majorVersion string) string {
-	// Use the highest compatible major version to avoid unwanted upgrades.
-	// See the comment on workspaceModuleVersion.
-	if majorVersion == "v0" {
-		majorVersion = "v1"
-	}
-	return majorVersion + workspaceModuleVersion
-}
